if SCrafts return end

rawset(_G, "SCrafts", true)

if not C2SBarEdits
	print("\x82".."NOTICE: ".."\x80".."S-Crafts are dependent on C2SBarEdits, which must be loaded before this mod in order to be utilized.")
	return
end

------		[READ ME]		------
/*
S-Crafts are special, powerful skills from the Kiseki/Trails game series, the equivalent of super attacks,
and have very special mechanics in order to be made properly.

S-Crafts should use the following flags:

		cost = 0,					--We have to circumvent standard skill costs in order to read the user's current CP amount on use.
		forcetp = true,				--
		notpgain = true,			--
		scraft = true,				--This flag confirms the skill as being an S-Craft for damage and cost handling.
		nometronome = true,			--
		norandom = true,			--

Optional Flags:

		strengthoverride = true, 	--This allows SP/TP skills to use STR instead of MAG for damage calculations.

S-Crafts should have the following nocast function:

		nocast = 	function(targets, n, mo)
				if not charStats[mo.stats].tpbar or charStats[mo.stats].tpmode ~= "craft points"
					BTL_logMessage(targets[1].battlen, "You can't use this skill.")
					return true
				end
				if charStats[mo.stats].tpspuse
					if mo.tp < 100
						BTL_logMessage(targets[1].battlen, "Insufficient Craft Points")
						return true
					end
				else
					if mo.sp < 100
						BTL_logMessage(targets[1].battlen, "Insufficient Craft Points")
						return true
					end
				end
			end,

S-Crafts require 100 TP to cast, but will zero-out our current TP upon use. The more TP we have when we cast an
S-Craft, the stronger its effect will be. It's also a Kiseki standard for some S-Crafts to have additional effects
when cast at 200 CP (CP, Craft Points, is what we're using TP as a stand-in for.)

-From this point on, we'll refer to TP as being CP unless directly meaning base, unmodified TP behavior-

This addon gives an alternate behavior to TP to make it work as faithfully as possible to the Kiseki CP system.
Characters who use CP have the following TP behavior changes:
	*Fixed 200 CP capacity
	*CP is completely depleted upon death
	*CP is retained between fights instead of resetting at the end of battle

To make a character use the CP system instead of standard TP, they require the following TP Mode:
		tpmode = "craft points",

It should also be noted that defining a character's S-Crafts in their charStats will do nothing if they are not
using the CP system, though the character is not required to have SP functionality (tpspuse charStats flag is optional).
*/

local function addSkill (mo, skill)
			
	local check = false
					
	if #mo.skills
		for s = 1,#mo.skills
		local x = mo.skills[s]
			if skill == x
				check = true
			end
		end
	end
										
	if #mo.passiveskills
		for s = 1,#mo.passiveskills
		local x = mo.passiveskills[s]
			if skill == x
				check = true
				else
			end
		end
	end
									
	if not check
		if attackDefs[skill].type & ATK_PASSIVE
			mo.passiveskills[#mo.passiveskills+1] = skill
		else	
			mo.skills[#mo.skills+1] = skill
		end
	end
end

local function removeSkill (mo, skill)
	
	local storenum = 0
				
	if #mo.skills
		for i = 1,#mo.skills
		local x = mo.skills[i]
			if skill == x
				storenum = i
			end
		end
	end
				
	if #mo.passiveskills
		for i = 1,#mo.passiveskills
			local x = mo.passiveskills[i]
			if skill == x
				storenum = i
			end
		end
	end
				
	if storenum
		if attackDefs[skill].type & ATK_PASSIVE
			table.remove(mo.passiveskills,storenum)
		else	
			table.remove(mo.skills,storenum)
		end
	end
end

-- Main S-Craft function --

local S_Craft = BTL_splitSkills	--S Crafts
BTL_splitSkills = function(v)
	if v and v.valid and v.stats and charStats[v.stats] and charStats[v.stats].persona and not v.enemy
		if charStats[v.stats].s_craft and charStats[v.stats].tpmode and charStats[v.stats].tpmode == "craft points" and v.level
			if server.gamemode == GM_VOIDRUN
				return
			end
			
			local scraft = 0
			local scraft2 = 0	--up to two S-Craft line slots. Kiseki usually gets 3, but we're already pushing it with "more skill slots" as it is.
			for k,j in ipairs(charStats[v.stats].s_craft or {})
				if j

					if v.level >= j[2]
						-- check skill flags
						if not attackDefs[j[1]].scraft		--Must work with the S Craft triggerdamage passive.
						or not attackDefs[j[1]].forcetp		--Not super important, but I'll enforce this for resource communication.
						or not attackDefs[j[1]].notpgain	--I'll lift this when C2S makes TP gain more open source.
						or not attackDefs[j[1]].nometronome	--Probably a good idea not to let Metronome pull these.
						or not attackDefs[j[1]].norandom	--Ditto above.
						or not attackDefs[j[1]].norogue		--It'd be fine if not for Emerl and Kirby.
						or not attackDefs[j[1]].nocast		--I can't check the exact function here but you better have it anyway.
						--or not attackDefs[j[1]].noenemycast	--We should be able to handle enemies using S-crafts safely now.
							print("\x82".."Warning: ".."\x80".."Loaded skill is not a valid S-Craft. Please be sure to only put S-Craft flagged skills in the S-Craft slot and that they meet the required criteria.")
							playSound(v.battlen, sfx_not)
							break
						end
						-- add skill
						if j[3] >= 2
							scraft2 = j[1]
						else
							scraft = j[1]
						end
					end
				end
			end
			if scraft then v.skills[#v.skills+1] = scraft end
			if scraft2 then v.skills[#v.skills+1] = scraft2 end

			-- autosort skills
			table.sort(v.skills, attackSortFunc)
			
			if v.passiveskills	--Everyone using S Crafts needs this passive.
				table.insert(v.passiveskills, "scraft_boost")
			else
				table.insert(v.skills, "scraft_boost")
			end
		end
	end
	S_Craft(v)
end

-- Enemy S-Craft function --

SRB2P_addHook("NextTurn", function(mo)
	if mo and mo.valid and mo.enemy and enemyList[mo.enemy] and mo.tpbar
	and (mo.tpmode == "craft points" or enemyList[mo.enemy].tpmode == "craft points") and (mo.s_craft or enemyList[mo.enemy].s_craft)
		local cp = 0
		if (mo.tpspuse or enemyList[mo.enemy].tpspuse)
			cp = mo.tp
		else
			cp = mo.sp
		end
		if cp >= 100 --they can and will ignore nocasts so we have to do this check.
			if server.gamemode == GM_VOIDRUN	--redundant but still wouldn't be fair if players can't when enemies can
				return
			end
			dprint("Enemy "..mo.name.." has "..cp.." CP, adding S-Crafts.")
			local scraft = 0
			local scraft2 = 0	--up to two S-Craft line slots. Kiseki usually gets 3, but we're already pushing it with "more skill slots" as it is.
			for k,j in ipairs((mo.s_craft or enemyList[mo.enemy].s_craft) or {})
				if j
					local level = 1
					if mo.level and mo.level > 1 then level = mo.level end	--in case some one makes an enemy with no level, if that's possible.
					if level >= j[2]
						-- check skill flags
						if not attackDefs[j[1]].scraft		--Must work with the S Craft triggerdamage passive.
						or not attackDefs[j[1]].forcetp		--Not super important, but I'll enforce this for resource communication.
						or not attackDefs[j[1]].notpgain	--I'll lift this when C2S makes TP gain more open source.
						or not attackDefs[j[1]].nometronome	--Probably a good idea not to let Metronome pull these.
						or not attackDefs[j[1]].norandom	--Ditto above.
						or not attackDefs[j[1]].norogue		--It'd be fine if not for Emerl and Kirby.
						or not attackDefs[j[1]].nocast		--I can't check the exact function here but you better have it anyway.
						--or not attackDefs[j[1]].noenemycast	--We should be able to handle enemies using S-crafts safely now.
							print("\x82".."Warning: ".."\x80".."Loaded skill is not a valid S-Craft. Please be sure to only put S-Craft flagged skills in the S-Craft slot and that they meet the required criteria.")
							playSound(mo.battlen, sfx_not)
							break
						end
						-- add skill
						if j[3] >= 2
							scraft2 = j[1]
						else
							scraft = j[1]
						end
					end
				end
			end
			if scraft then addSkill(mo, scraft) end
			if scraft2 then addSkill(mo, scraft2) end

			-- autosort skills
			table.sort(mo.skills, attackSortFunc)
			addSkill(mo, "scraft_boost")	--Everyone using S Crafts needs this passive.
		else	--shouldn't be casting S-Crafts if the requirements are not properly met
			dprint("Enemy "..mo.name.." has "..cp.." CP, removing S-Crafts.")
			if #mo.skills
				local gone = {}
				for i = 1,#mo.skills
				local x = mo.skills[i]
					if attackDefs[x].scraft
						gone[#gone+1] = i
					end
				end
				for i = 1,#gone
					table.remove(mo.skills,gone[i])
				end
			end
		end
	end
end)


attackDefs["scraft_boost"] = {
		name = "S-Craft Boost",
		type = ATK_PASSIVE,
		desc = "Damage booster for S-Craft skills. \nThe more CP you have, the harder \nS-Crafts will hit!",
		passive = PSV_TRIGGERDAMAGE,
		anim = 	function(mo, target, dmg, crit)
					if mo.attack.scraft
						dprint(mo.name.." attempting S-Craft.")
						if (mo.cpuse or (mo.stats and charStats[mo.stats].tpmode and charStats[mo.stats].tpmode == "craft points") or (mo.enemy and enemyList[mo.enemy].tpmode and enemyList[mo.enemy].tpmode == "craft points"))
						and (((mo.tpspuse or (mo.stats and charStats[mo.stats].tpspuse) or (mo.enemy and enemyList[mo.enemy].tpspuse)) and mo.tp) or mo.sp)
							dprint(mo.name.." Has CP, checking quantity.")
							local warn = false
							if (mo.tpspuse or (mo.stats and charStats[mo.stats].tpspuse) or (mo.enemy and enemyList[mo.enemy].tpspuse))
								dprint(mo.name.." Has "..mo.tp.." CP.")
								if mo.tp >= 200
									dmg = $*25/10
								elseif mo.tp < 100
									dmg = 0
									warn = true
								else
									dmg = $*(mo.tp)/100
								end
								--mo.tp = 0
							else
								dprint(mo.name.." Has "..mo.sp.." CP.")
								if mo.sp >= 200
									dmg = $*25/10
								elseif mo.sp < 100
									dmg = 0
									warn = true
								else
									dmg = $*(mo.sp)/100
								end
								--mo.sp = 0
							end
							if warn
								BTL_logMessage(mo.battlen, "Invalid S-craft use! Damage set to 0!")
								print("\x82".."Warning: ".."\x80".."Invalid S-Craft use detected! Please be sure to use the provided nocast funtion in your S-Craft's attakdefs to prevent misuse.")
								playSound(mo.battlen, sfx_not)
								mo.setonemore = false
							end
						end
						if not (mo.cpuse or (mo.stats and charStats[mo.stats].tpmode and charStats[mo.stats].tpmode == "craft points") or (mo.enemy and enemyList[mo.enemy].tpmode and enemyList[mo.enemy].tpmode == "craft points"))
							BTL_logMessage(mo.battlen, "Invalid S-craft use! Damage set to 0!")
							print("\x82".."Warning: ".."\x80".."Attempted use of S-Craft with an invalid character type detected! Please be sure to have tpbar set to true in your charStats, and have your tpmode set to 'craft points' or set cpuse to true in there as well.")
							playSound(mo.battlen, sfx_not)
							mo.sp = 0
							dmg = 0
							mo.setonemore = false
						end
						return dmg
					end
				end,
	}

-- CP capacity and death penalty --

	
local function craftpoints(mo)
	if not mo or not mo.valid or not (mo.tpbar or (mo.stats and charStats[mo.stats] and charStats[mo.stats].tpbar)) return end
	if server.gamemode == GM_VOIDRUN return end --not in void run please
	if not mo.cpuse and not (mo.stats and charStats[mo.stats] and charStats[mo.stats].tpmode and charStats[mo.stats].tpmode == "craft points") return end	--not a craftpoints user
	if not mo.cpuse then mo.cpuse = true end
	if (mo.stats and charStats[mo.stats].tpspuse) and not mo.tpuse then mo.tpuse = true end
	local btl = server.P_BattleStatus[mo.battlen]
	if btl and btl.running 
		if mo.tpspuse
			if not mo.hp and mo.tp then mo.tp = max(0, $-4) end
		else
			if not mo.hp and mo.sp then mo.sp = max(0, $-4) end
		end
		if btl.battlestate == BS_ENDTURN
			if mo.attack and mo.attack.scraft	--In case we miss
				if mo.tpspuse
					if mo.tp then mo.tp = 0 end
				else
					if mo.sp then mo.sp = 0 end
				end
				if mo.enemy	--because I guess better baton pass skips the nextturn hook sometimes?!
					dprint("Enemy "..mo.name.." S-Craft finished, removing S-Crafts.")
					if #mo.skills
						local gone = {}
						for i = 1,#mo.skills
						local x = mo.skills[i]
							if attackDefs[x].scraft
								gone[#gone+1] = i
							end
						end
						for i = 1,#gone
							table.remove(mo.skills,gone[i])
						end
					end
				end
			end
		end
	end
end

addHook("MobjThinker", craftpoints, MT_PFIGHTER)


SRB2P_addHook("EnemyInit", function(mo)
	if (mo.enemy and enemyList[mo.enemy].tpmode and enemyList[mo.enemy].tpmode == "craft points") and (mo.tpbaar or (mo.enemy and enemyList[mo.enemy].tpbar)) and not mo.cpuse
		mo.cpuse = true
	end
	if mo.enemy == "clonechar" or mo.enemy == "clonechar-hard"
		if mo.clonedfrom and (mo.clonedfrom.cpuse or charStats[mo.clonedfrom.stats].tpmode and charStats[mo.clonefrom.stats].tpmode == "craft points") and not mo.cpuse
			local charDefs = charStats[mo.clonedfrom.stats]
			mo.tp = mo.clonedfrom.tp
			mo.maxtp = mo.clonedfrom.maxtp
			mo.cpuse = true
			if charStats[mo.clonedfrom.stats].scraft and not mo.scraft
				mo.s_craft = charDefs.scraft
			end
		end
	end
end)

-- The actual TP mode itself --

tpModes["craft points"] = {
		name = "CRAFT POINTS", -- for character select (should be in all caps for consistency)
		maxtp = 200, -- how much max tp enemies start with
		tp = 50, -- how much tp enemies start with
		canincreasemax = false, -- can you increase the max with stuff like wardrings?
		canusetea = 2, -- can you use tea in the overworld? 0/nil = no, 1 = limited use, 2 = unlimited use
		teascale = 10,
		gaintponstatus = 0, -- do we gain tp when inflicting and getting inflicted by status? 0/nil = no, 1 = inflicting, 2 = inflicted, 3 = both
		dungstartfunc = function(mo) -- function for when the dungeon starts
			mo.cpuse = true -- just in case
			if mo.tpbar
			and not mo.tpspuse
			and mo.maxsp <= 0 -- only set it if we don't already have it
				mo.realsp = 200
				mo.maxsp = 200
				mo.sp = 0
			elseif mo.tpbar
			and mo.tpspuse
			and mo.maxtp <= 0 -- only set it if we don't already have it
				mo.maxtp = 200
				mo.tp = 0
			end
		end,
		dungeonfunc = function(mo) -- function that runs constantly outside of battle
			if server.P_DungeonStatus.gameoverfade -- reset our tp riiight before we fade back in
				if charStats[mo.stats].tpspuse
					mo.tp = 0
				else
					mo.sp = 0
				end
			end
		end,
		pvpstartfunc = function(mo, targets, hittargets, timer) -- function when battle starts in pvp/vision quests
			if mo.tpspuse
				mo.tp = 50
			else
				mo.sp = 50
			end
		end,
		gainfunc = tpModes["default"].gainfunc,
	}
	
-- Parity Check with characters still using the old implementation --
addHook("ThinkFrame", function()
	for k in pairs(charStats)
		local character = charStats[k]
		if character.tpbar
		and character.cpuse
		and character.tpmode ~= "craft points"
			character.tpmode = "craft points"
		end
	end
	for k in pairs(enemyList)
		local enemy = enemyList[k]
		if enemy.tpbar
		and enemy.cpuse
		and enemy.tpmode ~= "craft points"
			enemy.tpmode = "craft points"
		end
	end
end)
/*
-- Bar Edits TP items rewrite --


itemDefs["diettea"] = {
		name = "Diet Tea",
		desc = "Slight TP recovery to an ally.\nStart next battle at 15% TP\nwhen used outside of battle.",
		attack = "heal_tp_0",
		cost = 900,
		checkfunc = function(user, mo)	-- so the level cap can't go too high
			if server.oldtp
				dprint("New TP isn't active!")
				return true 
			end
			if not charStats[mo.stats].tpbar
				dprint("TP isn't active!")
				return true
			end
			if mo.teatp and not charStats[mo.stats].cpuse
				dprint("Tea has already been used!")
				return true
			end
			if (mo.tpspuse and mo.tp >= mo.maxtp) or (mo.sp >= mo.maxsp and not mo.tpspuse)
				dprint("TP is already at max!")
				return true
			end
		end,
		menufunc = 	function(user, mo)	-- function to run when using the item from dungeon menu specifically
			-- user: player mo in dungeon
			-- mo: actual entity it was used on (will be a table if it's multi target)
			S_StartSound(nil, sfx_heal, user.player)
			if charStats[mo.stats].cpuse
				if mo.tpspuse
					mo.tp = min(mo.maxtp, $+5)
					mo.teatp = mo.tp
				else
					mo.sp = min(mo.maxsp, $+5)
					mo.teatp = mo.sp
				end
			else
				if mo.tpspuse
					mo.teatp = mo.maxtp*15/100
					mo.tp = mo.teatp
				else
					mo.teatp = mo.maxsp*15/100
					mo.sp = mo.teatp
				end
			end
		end,
		rarity = 3,
	}

itemDefs["lucrativetea"] = {
		name = "Lucrative Tea",
		desc = "Moderate TP recovery to an ally.\nStart next battle at 25% TP\nwhen used outside of battle.",
		attack = "heal_tp_1",
		cost = 900,
		checkfunc = function(user, mo)	-- so the level cap can't go too high
			if server.oldtp
				dprint("New TP isn't active!")
				return true 
			end
			if not charStats[mo.stats].tpbar
				dprint("TP isn't active!")
				return true
			end
			if mo.teatp and not charStats[mo.stats].cpuse
				dprint("Tea has already been used!")
				return true
			end
			if (mo.tpspuse and mo.tp >= mo.maxtp) or (mo.sp >= mo.maxsp and not mo.tpspuse)
				dprint("TP is already at max!")
				return true
			end
		end,
		menufunc = 	function(user, mo)	-- function to run when using the item from dungeon menu specifically
			-- user: player mo in dungeon
			-- mo: actual entity it was used on (will be a table if it's multi target)
			S_StartSound(nil, sfx_heal, user.player)
			if charStats[mo.stats].cpuse
				if mo.tpspuse
					mo.tp = min(mo.maxtp, $+10)
					mo.teatp = mo.tp
				else
					mo.sp = min(mo.maxsp, $+10)
					mo.teatp = mo.sp
				end
			else
				if mo.tpspuse
					mo.teatp = mo.maxtp/4
					mo.tp = mo.teatp
				else
					mo.teatp = mo.maxsp/4
					mo.sp = mo.teatp
				end
			end
		end,
		rarity = 3,
	}

itemDefs["volatiletea"] = {
		name = "Volatile Tea",
		desc = "High TP recovery to an ally.\nStart next battle at 50% TP\nwhen used outside of battle.",
		attack = "heal_tp_2",
		cost = 1750,
		checkfunc = function(user, mo)	-- so the level cap can't go too high
			if server.oldtp
				dprint("New TP isn't active!")
				return true 
			end
			if not charStats[mo.stats].tpbar
				dprint("TP isn't active!")
				return true
			end
			if mo.teatp and not charStats[mo.stats].cpuse
				dprint("Tea has already been used!")
				return true
			end
			if (mo.tpspuse and mo.tp >= mo.maxtp) or (mo.sp >= mo.maxsp and not mo.tpspuse)
				dprint("TP is already at max!")
				return true
			end
		end,
		menufunc = 	function(user, mo)	-- function to run when using the item from dungeon menu specifically
			-- user: player mo in dungeon
			-- mo: actual entity it was used on (will be a table if it's multi target)
			S_StartSound(nil, sfx_heal, user.player)
			if charStats[mo.stats].cpuse
				if mo.tpspuse
					mo.tp = min(mo.maxtp, $+20)
					mo.teatp = mo.tp
				else
					mo.sp = min(mo.maxsp, $+20)
					mo.teatp = mo.sp
				end
			else
				if mo.tpspuse
					mo.teatp = mo.maxtp/2
					mo.tp = mo.teatp
				else
					mo.teatp = mo.maxsp/2
					mo.sp = mo.teatp
				end
			end
		end,
		rarity = 4,
	}

itemDefs["explosivetea"] = {
		name = "Explosive Tea",
		desc = "Full TP recovery to an ally.\nStart next battle at 75% TP\nwhen used outside of battle.",
		attack = "heal_tp_3",
		cost = 4000,
		checkfunc = function(user, mo)	-- so the level cap can't go too high
			if server.oldtp
				dprint("New TP isn't active!")
				return true 
			end
			if not charStats[mo.stats].tpbar
				dprint("TP isn't active!")
				return true
			end
			if mo.teatp and not charStats[mo.stats].cpuse
				dprint("Tea has already been used!")
				return true
			end
			if (mo.tpspuse and mo.tp >= mo.maxtp) or (mo.sp >= mo.maxsp and not mo.tpspuse)
				dprint("TP is already at max!")
				return true
			end
		end,
		menufunc = 	function(user, mo)	-- function to run when using the item from dungeon menu specifically
			-- user: player mo in dungeon
			-- mo: actual entity it was used on (will be a table if it's multi target)
			S_StartSound(nil, sfx_heal, user.player)
			if charStats[mo.stats].cpuse
				if mo.tpspuse
					mo.tp = min(mo.maxtp, $+40)
					mo.teatp = mo.tp
				else
					mo.sp = min(mo.maxsp, $+40)
					mo.teatp = mo.sp
				end
			else
				if mo.tpspuse
					mo.teatp = mo.maxtp*3/4
					mo.tp = mo.teatp
				else
					mo.teatp = mo.maxsp*3/4
					mo.sp = mo.teatp
				end
			end
		end,
		rarity = 5,
	}
*/













